import os

from .client import client
from .context import Context, Parameter


class ZafiraState:
    """
    Zafira global state representation

    The attributes are:

    .. attribute:: zc
        Zafira API client used for most of activities in listener

    .. attribute:: is_enabled
        Flag, if False - Zafira connector is shutting down

    .. attribute:: service_url
        Zafira web service url

    .. attribute:: access_token
        Long-term access user token generated by Zafira user

    .. attribute:: job_name
        name of CI job where testrun is running

    .. attribute:: artifact_log_name
        name of entry of artifacts

    .. attribute:: artifact_expires_in_default_time
        expiration time of artifact (e.g. screenshot from Amazon)

    .. attribute:: circle_ci_client
        client of CIRCLECI API used for any job info

    .. attribute:: suite_name
        name of test suite

    .. attribute:: job_url
        url of CIRCLECI job

    .. attribute:: user
        user payload for Zafira web service

    .. attribute:: job
        job payload for Zafira web service

    .. attribute:: test_suite
        test suite payload for Zafira web service

    .. attribute:: refresh_token
        used for authentication of user in Zafira

    .. attribute:: test_case
        test case payload for Zafira web service

    .. attribute:: test_run
        test run payload for Zafira web service

    .. attribute:: test
        test payload for Zafira web service

    .. attribute:: ci_test_id
        used for correlation_id in Rabbitmq logging handler, the second part of it

    .. attribute:: ci_run_id
        used for correlation_id in Rabbitmq logging handler, the first part of it

    .. attribute:: skip_reason
        retrieves info from pytest tag, used for log in Zafira UI

    .. attribute:: class_name
        name of test class or feature for grouping by test classes/features in Zafira UI

    .. attribute:: test_name
        name of test used for Zafira UI

    .. attribute:: package
        name of test package grouping by test packages/testruns in Zafira UI

    .. attribute:: MAX_LENGTH_OF_WORKITEM
        maximal length of work item visible on Zafira UI

    .. attribute:: zafira_project
        project filter on Zafira UI

    """

    MAX_LENGTH_OF_WORKITEM = 46

    CONFIG = None

    INSTANCE = None

    def __new__(cls):
        if not cls.INSTANCE:
            cls.INSTANCE = super(ZafiraState, cls).__new__(cls)
        return cls.INSTANCE

    def __init__(self):
        self.zc = client

        self.is_enabled = eval(Context.get(Parameter.ZAFIRA_ENABLED))
        self.service_url = Context.get(Parameter.SERVICE_URL)
        self.access_token = Context.get(Parameter.ACCESS_TOKEN)
        self.zafira_project = Context.get(Parameter.ZAFIRA_PROJECT)

        self.suite_name = str(os.environ.get('JOB_NAME'))
        self.job_url = str(os.environ.get('JOB_URL'))

        self.refresh_token = None

        self.test_run_id = None
        self.test_id = None

        self.user = None
        self.job = None
        self.test_suite = None
        self.test_case = None
        self.test_run = None
        self.test = None

        self.ci_test_id = ''
        self.ci_run_id = ''

        self.skip_reason = None

        self.class_name = None
        self.test_name = None
        self.package = None



__all__ = ['zafira_state', 'client', 'resource_constants', 'resources', 'plugin', 'listener',
           'connector_obj', 'PyTestZafiraListener', 'handler', 'ZebrunnerRestHandler']

from .plugin import connector_obj
from .listener import PyTestZafiraListener
from .handler import ZebrunnerRestHandler



import configparser
import logging
from os import getcwd, environ
from enum import Enum

from .exceptions import ConfigError

CONFIG_FILE_PATH = getcwd() + '/zafira_properties.ini'

LOGGER = logging.getLogger('zebrunner')


class Context:

    @staticmethod
    def get(parameter):
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE_PATH)
        LOGGER.debug('Acquiring property {}'.format(parameter.value))
        return config.get('config', parameter.value)

    @staticmethod
    def get_list(parameter):
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE_PATH)
        LOGGER.debug('Acquiring property {}'.format(parameter.value))
        return config.get('config', parameter.value).split(',')


class Parameter(Enum):
    SERVICE_URL = 'service_url'
    ACCESS_TOKEN = 'access_token'
    ZAFIRA_ENABLED = 'zafira_enabled'
    ZAFIRA_PROJECT = 'zafira_project'
    DRIVER_INSTANCE_NAME = 'driver_instance_name'
    SUITE = 'suite'
    BUILD = 'build'
    ENV = 'env'
    TEST_OWNERS = 'test_owners'
    TEST_RUN_ARTIFACT = 'test_run_artifact'
    TEST_RUN_ARTIFACT_URL = 'test_run_artifact_url'


def get_env_var(env_var_key):
    """
    Getter for environment variable keys.  Uses `os.environ.get(KEYNAME)`
    :param env_key:  Type string.  Key name of environment variable to get.
    :return:  Value of the environment variable.
    """
    env_var_value = environ.get(env_var_key)
    if env_var_value is None:
        LOGGER.error('ENV var missing: [{0}], please set this variable'.format(env_var_key))
        raise ConfigError("Env variable {0} is mandatory, please set this variable".format(env_var_key))
    else:
        LOGGER.debug('ENV variable is: [{0}]:[{1}]'.format(env_var_key, env_var_value))
    return env_var_value



import datetime
import time
from copy import deepcopy

from zebrunnerpy.context import Parameter, Context
from .api_request import APIRequest
from .resources import test_v1, test_result_v1, refresh_token, test_log, test_artifact

from .resources import test_run_v1


class ZafiraClient:

    TEST_RUNS_PATH_V1 = '/api/reporting/v1/test-runs?projectKey={}'
    REFRESH_TOKEN_PATH = "/api/iam/v1/auth/refresh"
    TESTS_PATH_V1 = '/api/reporting/v1/test-runs/{}/tests'
    TESTS_FINISH_PATH_V1 = '/api/reporting/v1/test-runs/{}/tests/{}'
    TEST_RUNS_FINISH_PATH_V1 = '/api/reporting/v1/test-runs/{}'
    TEST_LOGS_PATH = '/api/reporting/v1/test-runs/{}/logs'
    TEST_SCREENSHOTS_PATH = '/api/reporting/v1/test-runs/{}/tests/{}/screenshots'
    TEST_RUN_ARTIFACT_V1 = '/api/reporting/v1/test-runs/{}/artifact-refs'

    INSTANCE = None

    def __new__(cls, service_url):
        if not cls.INSTANCE:
            cls.INSTANCE = super(ZafiraClient, cls).__new__(cls)
        return cls.INSTANCE

    def __init__(self, service_url):
        self.api = APIRequest(service_url)
        self.auth_token = ''

    def start_test_run(self, started_at, framework, project):
        body = deepcopy(test_run_v1)
        body["name"] = Context.get(Parameter.SUITE)
        started_at = started_at.replace(tzinfo=datetime.timezone.utc)
        body["startedAt"] = started_at.isoformat()
        body["framework"] = framework
        body["config"]["env"] = Context.get(Parameter.ENV)
        body["config"]["appVersion"] = Context.get(Parameter.BUILD)
        return self.api.send_post(ZafiraClient.TEST_RUNS_PATH_V1.format(project), body,
                                  headers=self.init_auth_headers(),
                                  default_err_msg="Unable to start test run")

    def finish_test_run(self, test_run):
        test_run["endedAt"] = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()
        return self.api.send_put(ZafiraClient.TEST_RUNS_FINISH_PATH_V1.format(test_run["id"]), test_run,
                                  headers=self.init_auth_headers(), default_err_msg="Unable to finish test run")

    def start_test(self, uid, test_run_id, test_name, maintainer, started_at, class_name):
        body = deepcopy(test_v1)
        body["uuid"] = uid
        body["name"] = test_name
        body["maintainer"] = maintainer
        started_at = started_at.replace(tzinfo=datetime.timezone.utc)
        body["startedAt"] = started_at.isoformat()
        body["className"] = class_name
        return self.api.send_post(ZafiraClient.TESTS_PATH_V1.format(test_run_id), body, headers=self.init_auth_headers(),
                                  default_err_msg="Unable to start test")

    def finish_test(self, test_run_id, test_id, test):
        body = deepcopy(test_result_v1)
        body["id"] = test_id
        body["uuid"] = test["uuid"]
        body["name"] = test["name"]
        body["startedAt"] = test["startedAt"]
        body["className"] = test["className"]
        body["methodName"] = test["methodName"]
        body["result"] = test["result"]
        body["reason"] = test.get("reason") or "passed"
        body["endedAt"] = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()
        return self.api.send_put(ZafiraClient.TESTS_FINISH_PATH_V1.format(test_run_id, test_id), body,
                                  headers=self.init_auth_headers(), default_err_msg="Unable to finish test")

    def push_log_record(self, test_run_id, test_id, record):
        body = deepcopy(test_log)
        list_entry = {
            "testId": test_id,
            "level": record.levelname,
            "timestamp": time.time() * 1000,
            "message": record.msg
        }
        body.append(list_entry)
        return self.api.send_post(ZafiraClient.TEST_LOGS_PATH.format(test_run_id), body, headers=self.init_auth_headers(), default_err_msg="Unable to send log entry")

    def push_screenshot(self, test_run_id, test_id, image):
        return self.api.send_post_screenshot(ZafiraClient.TEST_SCREENSHOTS_PATH.format(test_run_id, test_id), image, headers=self.init_auth_headers_with_screenshot(), default_err_msg="Unable to send screenshot")

    def push_artifact(self, test_run_id):
        body = deepcopy(test_artifact)
        body["name"] = Context.get(Parameter.TEST_RUN_ARTIFACT)
        body["url"] = Context.get(Parameter.TEST_RUN_ARTIFACT_URL)
        return self.api.send_post(ZafiraClient.TEST_RUN_ARTIFACT_V1.format(test_run_id), body, headers=self.init_auth_headers(), default_err_msg="Unable to attach testrun artifact")

    def refresh_token(self, token):
        refresh_token["refreshToken"] = token
        return self.api.send_post_without_authorization(ZafiraClient.REFRESH_TOKEN_PATH, refresh_token,
                                                        default_err_msg="Unable to refresh token")

    def init_auth_headers(self):
        return {"Authorization": "Bearer " + self.auth_token}

    def init_auth_headers_with_screenshot(self):
        return {"Content-type": "image/png", "Authorization": "Bearer " + self.auth_token}


client = ZafiraClient(Context.get(Parameter.SERVICE_URL))



from .zafira_state import ZafiraState


class PytestZafiraConnector:
    """
    An instance of Zafira connector with generic Zafira state,
    Can be instantiated as pytest connector,
    or behave connector according to a running context
    """
    def __init__(self):
        self.state = ZafiraState()


connector_obj = PytestZafiraConnector()



refresh_token = {
  "refreshToken": "REFRESH_TOKEN"
}

test_run_v1 = {
  "name": "Test NG run",
  "startedAt": "2019-12-03T10:15:30+01:00",
  "framework": "testng",
  "config": {
    "env": "value",
    "appVersion": "value",
    "platform": "value"
  }
}


test_v1 = {
  "uuid": "1512351235-3215-3125-3215531",
  "name": "Test ABC",
  "className": "com.test.MyTests",
  "methodName": "featureTest()",
  "startedAt": "2007-12-03T10:15:30Z",
  "maintainer": "johndoe@zebrunner.com"
}


test_result_v1 = {
  "result": "failed",
  "endedAt": "2007-12-03T10:15:30Z"
}


test_log = []

test_artifact = {
  "name": "jenkins-log-20200604.tar.gz",
  "url": "https://ci.organization.org/pipelines/jenkins-log- 20200604.tar.gz"
}


import logging

from .plugin import connector_obj


class ZebrunnerRestHandler(logging.Handler):

    def __init__(self):
        logging.Handler.__init__(self, level=10)
        self.state = connector_obj.state

    def emit(self, record):
        test_run_id = self.state.test_run.json()["id"]
        test_id = self.state.test["id"]
        self.state.zc.push_log_record(test_run_id, test_id, record)



import logging

import requests


class APIRequest:

    LOGGER = logging.getLogger('zebrunner')

    def __init__(self, base_url):
        self.base_url = base_url

    # HTTP methods
    def send_post(self, endpoint, body=None, headers=None, default_err_msg=None):
        url = self.base_url + endpoint
        self.LOGGER.debug('POST {} \n Headers:{} \n Payload: {}'.format(url, headers, body))
        try:
            resp = requests.post(url, json=body, headers=headers)
        except Exception as e:
            self.LOGGER.error(default_err_msg, e)
        return self.__verify_response(resp)

    def send_post_screenshot(self, endpoint, body=None, headers=None, default_err_msg=None):
        url = self.base_url + endpoint
        self.LOGGER.debug('POST {} \n Headers:{} \n Payload: {}'.format(url, headers, body))
        try:
            resp = requests.post(url, body, headers=headers)
        except Exception as e:
            self.LOGGER.error(default_err_msg, e)
        return self.__verify_response(resp)

    def send_post_without_authorization(self, endpoint, body=None, default_err_msg=None):
        url = self.base_url + endpoint
        self.LOGGER.debug('POST {} \n Payload: {}'.format(url, body))
        try:
            resp = requests.post(url, json=body)
        except Exception as e:
            self.LOGGER.error(default_err_msg, e)
        return self.__verify_response(resp)

    def send_get(self, endpoint, headers=None, default_err_msg=None):
        url = self.base_url + endpoint
        self.LOGGER.debug('POST {} \n Headers:{} \n Payload: {}'.format(url, headers, body))
        try:
            resp = requests.get(url=url, headers=headers)
        except Exception as e:
            self.LOGGER.error(default_err_msg, e)
        return self.__verify_response(resp)

    def send_delete(self, endpoint=None, body=None, headers=None, default_err_msg=None):
        url = self.base_url + endpoint
        self.LOGGER.debug('POST {} \n Headers:{} \n Payload: {}'.format(url, headers, body))
        try:
            resp = requests.delete(url, json=body, headers=headers)
        except Exception as e:
            self.LOGGER.error(default_err_msg, e)
        return self.__verify_response(resp)

    def send_put(self, endpoint=None, body=None, headers=None, default_err_msg=None):
        url = self.base_url + endpoint
        self.LOGGER.debug('POST {} \n Headers:{} \n Payload: {}'.format(url, headers, body))
        try:
            resp = requests.put(url, json=body, headers=headers)
        except Exception as e:
            self.LOGGER.error(default_err_msg, e)
        return self.__verify_response(resp)

    @staticmethod
    def __verify_response(resp):
        """Log and check API call. In case status code of response is not 200, will raise an HTTPException """
        if not str(resp.status_code).startswith('2'):
            raise Exception("HTTP call fails. Response status code {status_code}".format(status_code=resp.status_code))
        return resp



import datetime
import logging
import pytest
import uuid

from zebrunnerpy.context import Context, Parameter
from .resource_constants import TestStatus


class BaseZafiraListener:

    LOGGER = logging.getLogger('zebrunner')

    def __init__(self, state):
        self.state = state

        self.initialized = self.initialize_zafira()

    def initialize_zafira(self):
        enabled = False
        try:

            self.state.refresh_token = self.state.zc.refresh_token(self.state.access_token).json()
            self.state.zc.auth_token = self.state.refresh_token['authToken']
            self.LOGGER.info("Zafira is " + "available" if self.state.is_enabled else "unavailable")

            enabled = self.state.is_enabled
        except Exception as e:
            self.LOGGER.error("Unable to initialize connector: {}".format(e))
        return enabled

    def compose_package_name(self, path_entries_list):
        if len(path_entries_list) == 2:
            return path_entries_list[0]
        return path_entries_list[0] + '/' + self.compose_package_name(path_entries_list[1:])

    def add_artifact_to_test(self, test, artifact_name, artifact_link, expires_in=None):
        """
        Adds test artifact to test
        """
        try:
            self.state.zc.add_test_artifact_to_test(test["id"], artifact_link, artifact_name, expires_in)
        except Exception as e:
            self.LOGGER.error("Unable to add artifact to test correctly: {}".format(e))

    def on_test_success(self):
        """
        An abstract handler for posting successfull test onto Zafira UI
        """

    def on_test_failure(self, message):
        """
        An abstract handler for posting failed test onto Zafira UI
        :param message: error message from failing test
        """

    def on_test_skipped(self, message):
        """
        An abstract handler for posting skipped test onto Zafira UI
        :param message: skip reason
        """

    def add_work_item_to_test(self, test_id, work_item):
        if not self.state.is_enabled:
            return
        try:
            work_items = list()
            work_items.append(work_item if len(work_item) < self.state.MAX_LENGTH_OF_WORKITEM else 'Skipped')
            self.state.zc.create_test_work_items(test_id, work_items)
        except Exception as e:
            self.LOGGER.error("Unable to add work item: {}".format(e))


class PyTestZafiraListener(BaseZafiraListener):
    """
    Contains hook implementations
    """

    FRAMEWORK = 'pytest'

    def __init__(self, state):
        super().__init__(state)

    @pytest.hookimpl
    def pytest_sessionstart(self, session):
        """
        Setup-class handler, signs in user, creates a testsuite,
        testcase, job and registers testrun in Zafira
        """
        if not self.initialized:
            return
        try:
            self.state.test_run = self.state.zc.start_test_run(
                datetime.datetime.utcnow(),
                self.FRAMEWORK,
                self.state.zafira_project
            )
            self.state.test_run_id = self.state.test_run.json()["id"]
            # self.state.zc.push_artifact(self.state.test_run_id)
        except Exception as e:
            self.state.is_enabled = False
            self.LOGGER.error("Undefined error during test run registration! {}".format(e))

    @pytest.hookimpl
    def pytest_runtest_setup(self, item):
        """
        Setup handler, set up initial parameters for test,
        attaches to testsuite, registers and starts the test
        """
        if not self.state.is_enabled:
            return
        try:
            test_name = item.name
            class_name = item.nodeid.split('::')[1]
            uid = str(uuid.uuid4())
            maintainer = [marker for marker in item.own_markers if
                          marker.name in Context.get_list(Parameter.TEST_OWNERS)]
            maintainer = 'anonymous' if len(maintainer) == 0 else maintainer[0].name
            self.state.test = self.state.zc.start_test(
                uid,
                self.state.test_run_id,
                test_name,
                maintainer,
                datetime.datetime.utcnow(),
                class_name
            ).json()
            self.state.test_id = self.state.test["id"]
        except Exception as e:
            self.LOGGER.error("Undefined error during test case/method start! {}".format(e))

    @pytest.hookimpl
    def pytest_runtest_teardown(self, item):
        """
        Teardown handler. Finishes test, adds workitems if needed
        """
        if not self.state.is_enabled:
            return
        try:
            skip_mark = [marker for marker in item.own_markers if marker.name == 'skip']
            if skip_mark:
                test_name = item.name
                class_name = item.nodeid.split('::')[1]
                uid = str(uuid.uuid4())
                maintainer = [marker for marker in item.own_markers if
                              marker.name in Context.get_list(Parameter.TEST_OWNERS)]
                maintainer = 'anonymous' if len(maintainer) == 0 else maintainer[0].name
                self.state.test = self.state.zc.start_test(
                    uid,
                    self.state.test_run_id,
                    test_name,
                    maintainer,
                    datetime.datetime.utcnow(),
                    class_name
                ).json()
                self.state.test_id = self.state.test["id"]
                self.state.test["result"] = TestStatus.SKIPPED.value
                self.state.test["reason"] = skip_mark[0].kwargs["reason"]

            self.state.zc.finish_test(self.state.test_run_id, self.state.test_id, self.state.test)
        except Exception as e:
            self.LOGGER.error("Unable to finish test run correctly: {}".format(e))

    @pytest.hookimpl
    def pytest_runtest_logreport(self, report):
        """
        Set test status, stacktrace if needed
        :param report: info about test
        """
        if not self.state.is_enabled:
            return
        try:
            if report.when == 'setup':
                if report.failed:
                    self.on_test_failure(report)
            if report.when == 'call':
                test_result = report.outcome
                if test_result is 'passed':
                    self.on_test_success()
                elif test_result is 'failed':
                    self.on_test_failure(report)
                else:
                    self.on_test_skipped(report)

        except Exception as e:
            self.LOGGER.error("Unable to finish test correctly: {}".format(e))

    @pytest.hookimpl
    def pytest_sessionfinish(self, session, exitstatus):
        """
        Teardown-class handler, closes the testrun
        """
        if not self.state.is_enabled:
            return
        try:
            self.state.zc.finish_test_run(self.state.test_run.json())
        except Exception as e:
            self.LOGGER.error("Unable to finish test run correctly: {}".format(e))

    @pytest.hookimpl
    def pytest_runtest_makereport(self, item, call):
        if not self.on_exception(item, call):
            return

        self.LOGGER.debug('Exception occurs... '
                          'Trying to catch screenshot')

        test_run_id = self.state.test_run.json()["id"]
        test_id = self.state.test["id"]

        self.state.zc.push_screenshot(
            test_run_id,
            test_id,
            item.instance.driver.get_screenshot_as_png()
        )

    @staticmethod
    def on_exception(item, call):
        return hasattr(item.instance, Context.get(Parameter.DRIVER_INSTANCE_NAME)) and call.excinfo

    def on_test_success(self):
        self.state.test["result"] = TestStatus.PASSED.value

    def on_test_failure(self, message):
        self.state.test["result"] = TestStatus.FAILED.value
        self.state.test["reason"] = message.longreprtext

    def on_test_skipped(self, message):
        if not hasattr(message, 'wasxfail'):
            self.state.test['result'] = TestStatus.SKIPPED.value
            self.state.test['reason'] = message.longreprtext
        else:
            self.state.test['result'] = TestStatus.FAILED.value
            self.state.test['reason'] = message.wasxfail
